/**
 * A bone box is a complete UI element for puppet show. A bone box does the following:
 * 1) Shows the Bone name (Text Tag)
 * 2) Allows the user to turn on/off the use of this tracker
 * 3) Set the Bone Name
 * 4) Set the Rotation Type
 * 5) Set the range of the motion (both min and max)
 * 6) Calibrate the tracker
 */

class ConfigBox{
  // constants BLAFOOOOEEE
  final static float INTERP_AMOUNT = .3;
  final static float CALIB_SIZE = 40;
  final static float CALIB_TIME = 4.9;

  final static String BOX_OPEN = "Box Open";
  final static String BOX_CLOSED = "Box Closed";

  final static String TRACKER_ON = "Tracker On";
  final static String TRACKER_OFF = "Tracker Off";

  final static String ROT_PITCH = "pitch";
  final static String ROT_YAW = "yaw";
  final static String ROT_ROLL = "roll";
  final static String ROT_BINARY = "binary";

  final static String CALIB_OFF = "dear john";
  final static String CALIB_RUNNING = "white:";
  final static String CALIB_CAPTURE = "a curse word for you";

  final static String PARAM_X = "Fool, i am going to used x!";
  final static String PARAM_Y = "y is so much better yo";

  // UI Elements
  private UIElement[] elements;
  private int elementCount;
  private TextTag tag_title;
  private TextTag tag_trackerstatus;
  private TextTag tag_on;
  private TextTag tag_off;
  private TextTag tag_name;
  private TextTag tag_rotationtype;
  private TextTag tag_pitch;
  private TextTag tag_roll;
  private TextTag tag_yaw;
  private TextTag tag_binary;
  private TextTag tag_rangefrom;
  private TextTag tag_rangeto;
  private TextTag tag_calibrate;
  private TextTag tag_param;
  private TextTag tag_x;
  private TextTag tag_y;

  private TextField field_name;
  private TextField field_rangefrom;
  private TextField field_rangeto;

  // global draw members
  private Point drawLocation;                   // where we are drawing now
  private Point finalDrawLocation;              // where we want to draw at
  private Point openLocation;                   // where we go when we're open
  private Point closedLocation;                 // where we go when we're closed
  private PFont titleFont;                      // the font to use for the title
  private PFont tagFont;                        // the font to use for the tags
  private PFont textBoxFont;                    // the other font, BASdljdhgdjs
  private PImage bck;
  // interaction members
  private boolean bAllowInteraction;            // does this guy even get to do anything?

  // state members
  private String boxState;                      // open or closed, i guess
  private String trackerState;                  // on or off, pretty much
  private String rotationState;                 // the type of rotation selected
  private String paramState;                    // x or y!
  private String calibrationState;              // the state of the calibration
  private float calibrationCount;               // keeping time of the calibration

  // Network Stuff
  private PuppetServer network;

  // Color tracking stuff
  private PImage frame;                         // the image to track on, poo (fuck)
  private PuppetBlobTracker tracker;
  private ColorArea cArea;

  ////
  // Constructors!
  ////
  public ConfigBox(PFont titleFont, PFont tagFont, PFont textBoxFont, Point drawLocation, PuppetServer network, PuppetBlobTracker tracker, PImage bG){
    this.titleFont = titleFont;
    this.tagFont = tagFont;
    this.textBoxFont = textBoxFont;
    this.closedLocation = drawLocation.copy();
    this.openLocation = drawLocation.copy();
    this.openLocation.y -= 160;
    this.drawLocation = drawLocation.copy();
    this.finalDrawLocation = this.closedLocation;
    
    this.network = network;
    this.tracker = tracker;
    tracker.setDrawBlobs(true);

    this.cArea = new ColorArea(int(width / 2 - CALIB_SIZE), int(height / 2 - CALIB_SIZE), int(width / 2 + CALIB_SIZE), int(height / 2 + CALIB_SIZE));

    this.elements = new UIElement[30];
    this.elementCount = 0;

    this.resetState();
    this.buildBox();
    
    this.bck = bG;
  }

  ////
  // Reset all the states!
  ////
  private void resetState(){
    this.boxState = BOX_CLOSED;
    this.trackerState = TRACKER_OFF;
    this.rotationState = ROT_PITCH;
    this.calibrationState = CALIB_OFF;
    this.paramState = PARAM_X;
  }

  ////
  // Builds all the UIElements and such, poooo
  ////
  private void buildBox(){
    // allllllllllllllllll the tags
    tag_title = new TextTag(this.titleFont, new Point(0, 0), 200, 25);
    tag_title.setText("<NO BONE>");
    this.addElement(tag_title);

    tag_trackerstatus = new TextTag(this.tagFont, new Point(0, 24), 0, 0);
    tag_trackerstatus.setText("TRACKER STATUS");
    this.addElement(tag_trackerstatus);

    tag_on = new TextTag(this.tagFont, new Point(200, 24), 30, 12);
    tag_on.setText("ON");
    this.addElement(tag_on);

    tag_off = new TextTag(this.tagFont, new Point(275, 24), 30, 12);
    tag_off.setText("OFF");
    this.addElement(tag_off);

    tag_name = new TextTag(this.tagFont, new Point(0, 24 + (18 * 1)), 0, 0);
    tag_name.setText("BONE NAME");
    this.addElement(tag_name);
    
    tag_param = new TextTag(this.tagFont, new Point(0, 24 + (18 * 2)), 0, 0);
    tag_param.setText("PARAMTER");
    this.addElement(tag_param);
    
    tag_x = new TextTag(this.tagFont, new Point(200, 24 + (18 * 2)), 46, 12);
    tag_x.setText("X AXIS");
    this.addElement(tag_x);
    
    tag_y = new TextTag(this.tagFont, new Point(275, 24 + (18 * 2)), 46, 12);
    tag_y.setText("Y AXIS");
    this.addElement(tag_y);
    
    tag_rotationtype = new TextTag(this.tagFont, new Point(0, 24 + (18 * 3)), 0, 0);
    tag_rotationtype.setText("SEND");
    this.addElement(tag_rotationtype);

    tag_pitch = new TextTag(this.tagFont, new Point(90, 24 + (18 * 3)), 42, 12);
    tag_pitch.setText("PITCH");
    this.addElement(tag_pitch);

    tag_yaw = new TextTag(this.tagFont, new Point(160, 24 + (18 * 3)), 35, 12);
    tag_yaw.setText("YAW");
    this.addElement(tag_yaw);

    tag_roll = new TextTag(this.tagFont, new Point(210, 24 + (18 * 3)), 35, 12);
    tag_roll.setText("ROLL");
    this.addElement(tag_roll);
    
    tag_binary = new TextTag(this.tagFont, new Point(260, 24 + (18 * 3)), 54, 12);
    tag_binary.setText("BINARY");
    this.addElement(tag_binary);

    tag_rangefrom = new TextTag(this.tagFont, new Point(0, 24 + (18 * 4)), 0, 0);
    tag_rangefrom.setText("RANGE FROM");
    this.addElement(tag_rangefrom);

    tag_rangeto = new TextTag(this.tagFont, new Point(200, 24 + (18 * 4)), 0, 0);
    tag_rangeto.setText("TO");
    this.addElement(tag_rangeto);

    tag_calibrate = new TextTag(this.titleFont, new Point(0, 24 + (18 * 6)), 150, 24);
    tag_calibrate.setText("CALIBRATE");
    this.addElement(tag_calibrate);

    // and alll the fields
    field_name = new TextField(this.textBoxFont, new Point(126, 24 + (18 * 1)), 150, 10);
    field_name.setText("<no bone>");
    this.addElement(field_name);

    field_rangeto = new TextField(this.textBoxFont, new Point(240, 24 + (18 * 4)), 60, 10);
    field_rangeto.setText("1");
    this.addElement(field_rangeto);

    field_rangefrom = new TextField(this.textBoxFont, new Point(126, 24 + (18 * 4)), 60, 10);
    field_rangefrom.setText("-1");
    this.addElement(field_rangefrom);
  }

  ////
  // adds a ui element in
  ////
  private void addElement(UIElement e){
    this.elements[this.elementCount] = e;
    this.elementCount ++; 
  }

  public void tick(float dt){
    // tracking
    if(this.trackerState == TRACKER_ON){
      this.tracker.setInputImg(this.frame);
      this.tracker.tick(dt);

      if(this.network != null){
        String send = "|" + this.field_name.getText() + "|" + this.rotationState + "|";
        
        Blob blob = this.tracker.findLargestBlob();
        if(blob != null){
          float center;
          
          if(this.paramState == PARAM_X){
            center = this.tracker.getAverageCentroid('x');
          } else {
            center = this.tracker.getAverageCentroid('y');
          }
          
          //normalize
          center += 1.0;
          center /= 2.0;
          
          float bottom = float(this.field_rangefrom.getText());
          float top = float(this.field_rangeto.getText());
          //println("TOP AND BOTTOM "+bottom+ " " +top);
          float sendValue = bottom + ((top - bottom) * center);
          
          send = send + "" + sendValue;
          //println("final string is: "+send);
          if(this.rotationState != ConfigBox.ROT_BINARY) this.network.sendString(send);
          else {
            String openString = "|" + this.field_name.getText() + "|" + this.field_rangefrom.getText();
            this.network.sendString(openString);
            println(openString);
          }
          
        }else{
            String closeString = "|" + this.field_name.getText() + "|" + this.field_rangeto.getText();
            this.network.sendString(closeString);
            //println(closeString);
        }
      } 
    }

    // calibration
    if(this.calibrationState == CALIB_OFF){

    } 
    else if(this.calibrationState == CALIB_RUNNING){
      this.calibrationCount += dt;
      this.cArea.captureColor(this.frame);
      if(this.calibrationCount >= CALIB_TIME){
        this.calibrationState = CALIB_CAPTURE; 
      }
    }
    else if(this.calibrationState == CALIB_CAPTURE){
      this.calibrationState = CALIB_OFF;
      println("Roll in the color change");
      int tempMask = 0xFF<<24 | (cArea.getAverageColor() & 0xFFFFFF);
      this.tracker.setAverageColor(this.cArea.getAverageColor());
      this.tracker.setMaskColor(tempMask);
      
      this.tracker.setLuminosityThreshold(this.cArea.getAverageLuminosity() / 3.0);
    } 
  }

  public void draw(){
    this.drawLocation = this.drawLocation.interp(INTERP_AMOUNT, this.finalDrawLocation);
    noStroke();
    fill(0);

    if(this.trackerState == TRACKER_ON){
      pushMatrix();
      translate(20, 20);
    this.tracker.draw();
    popMatrix();
    }
    

    pushMatrix();
    translate(this.drawLocation.x, this.drawLocation.y);
    image(bck,-5,-30);

    drawCard();
    popMatrix();


    if(this.calibrationState == CALIB_RUNNING){
      this.cArea.draw();
    }
  }

  ////
  // Draws the whole card thing
  ////
  private void drawCard(){
    // bar
    noStroke();
   fill(0);
    rect(0, 4, 300, 5);

    noStroke();
    fill(0, 100, 200);

    if(this.trackerState == TRACKER_ON){
      rect(tag_on.getLocation().x - 2, tag_on.getLocation().y - 12, tag_on.getWidth(), tag_on.getHeight() + 2);
    } 
    else if(this.trackerState == TRACKER_OFF){
      rect(tag_off.getLocation().x - 2, tag_off.getLocation().y - 12, tag_off.getWidth(), tag_off.getHeight() + 2);
    }

    if(this.rotationState == ROT_PITCH){
      rect(tag_pitch.getLocation().x - 2, tag_pitch.getLocation().y - 12, tag_pitch.getWidth() + 2, tag_pitch.getHeight() + 2);
    } 
    else if(this.rotationState == ROT_YAW){
      rect(tag_yaw.getLocation().x - 2, tag_yaw.getLocation().y - 12, tag_yaw.getWidth() + 2, tag_yaw.getHeight() + 2);
    } 
    else if(this.rotationState == ROT_ROLL){
      rect(tag_roll.getLocation().x - 2, tag_roll.getLocation().y - 12, tag_roll.getWidth() + 2, tag_roll.getHeight() + 2);
    }
    else if(this.rotationState == ROT_BINARY){
      rect(tag_binary.getLocation().x - 2, tag_binary.getLocation().y - 12, tag_binary.getWidth() + 2, tag_binary.getHeight() + 2);
    }
    
    if(this.paramState == PARAM_X){
      rect(tag_x.getLocation().x - 2, tag_x.getLocation().y - 12, tag_x.getWidth() + 2, tag_x.getHeight() + 2);
    }else if(this.paramState == PARAM_Y){
      rect(tag_y.getLocation().x - 2, tag_y.getLocation().y - 12, tag_y.getWidth() + 2, tag_y.getHeight() + 2);
    }

    // elements
    elementUpdate();

    //underline
    stroke(0);
    line(field_name.getLocation().x - 4, field_name.getLocation().y + 2, field_name.getLocation().x + 174, field_name.getLocation().y + 2);
    line(field_rangeto.getLocation().x - 4, field_rangeto.getLocation().y + 2, field_rangeto.getLocation().x + 60, field_rangeto.getLocation().y + 2);
    line(field_rangefrom.getLocation().x - 4, field_rangefrom.getLocation().y + 2, field_rangefrom.getLocation().x + 60, field_rangefrom.getLocation().y + 2); 
  }

  ////
  // update
  ////
  public void elementUpdate(){
    for(int i = 0; i < this.elementCount; i++){
      this.elements[i].tick(0);
      this.elements[i].draw();
    }
  }

  ////
  // input events
  ////
  void mousePressed(Point mouse){
    this.killFocus();
    
    Point normMouse = mouse.subPoint(this.drawLocation);
    if(this.tag_title.isHit(normMouse)){
      if(this.boxState == BOX_OPEN){
        this.boxState = BOX_CLOSED;
        this.finalDrawLocation = this.closedLocation;
        println("Box should be closing!");
      } 
      else {
        this.boxState = BOX_OPEN;
        this.finalDrawLocation = this.openLocation;
        println("Box should be opening!");
      }
    } else

    if(this.tag_on.isHit(normMouse)){
      this.trackerState = TRACKER_ON;
      println("tracker is on");
    }else
    if(this.tag_off.isHit(normMouse)){
      this.trackerState = TRACKER_OFF; 
      println("tracker is off");
    }else

    if(this.tag_pitch.isHit(normMouse)){
      this.rotationState = ROT_PITCH;
      println("pitch");
    }else
    if(this.tag_yaw.isHit(normMouse)){
      this.rotationState = ROT_YAW;
      println("yaw");
    }else
    if(this.tag_roll.isHit(normMouse)){
      this.rotationState = ROT_ROLL;
      println("roll");
    }else
    if(this.tag_binary.isHit(normMouse)){
      this.rotationState = ROT_BINARY;
      println("binary");
    }else
    if(this.tag_calibrate.isHit(normMouse)){
      println("Calibrate!");
      this.calibrationState = CALIB_RUNNING;
      this.calibrationCount = 0.0;
    }else
    if(this.tag_x.isHit(normMouse)){
      this.paramState = PARAM_X;
    }else
    if(this.tag_y.isHit(normMouse)){
      this.paramState = PARAM_Y;
    }

    if(this.field_name.isHit(normMouse)){
      this.field_name.setFocus(true);
      this.field_rangeto.setFocus(false);
      this.field_rangefrom.setFocus(false);
      println("name selected"); 
    } 
    else if(this.field_rangeto.isHit(normMouse)){
      this.field_name.setFocus(false);
      this.field_rangeto.setFocus(true);
      this.field_rangefrom.setFocus(false);
      println("rangeto selected"); 
    } 
    else if(this.field_rangefrom.isHit(normMouse)){
      this.field_name.setFocus(false);
      this.field_rangeto.setFocus(false);
      this.field_rangefrom.setFocus(true);
      println("rangefrom selected"); 
    }
  }
  
  
  void mouseReleased(Point mouse){
  
  }
  void keyPressed(char newChar){
    this.field_name.addCharacter(newChar);
    this.field_rangeto.addCharacter(newChar);
    this.field_rangefrom.addCharacter(newChar);

    if(newChar == 10){
      if(this.boxState == BOX_OPEN && this.field_name.bHasFocus()){
        this.tag_title.setText(this.field_name.getText().toUpperCase());
      }
    }
  }

  ////
  // Update the frame!
  ////
  public void setFrame(PImage frame){
    this.frame = frame; 
  }
  
  ////
  // Removes all the focus!
  ////
  public void killFocus(){
   for(int i = 0; i < this.elementCount; i++){
    this.elements[i].setFocus(false);
   } 
  }
}
